#!/usr/bin/env python
import sys
from qpy.compile import get_code
if __name__ == '__main__':
    file_name = sys.argv[1]
    code = get_code(open(file_name).read(), file_name)
    eval(code, dict(__name__='__main__', __file__=file_name))




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/test/utest_xml.py $
$Id: utest_xml.py 31098 2008-09-15 19:28:18Z dbinger $
"""
from sancho.utest import UTest
from qpy import xml, xml_quote
import sys

if sys.version.startswith("3"):
    unicode_string = str
else:
    from __builtin__ import unicode as unicode_string


class TestXml (UTest):

    def test_xml(self):
        u = xml()
        assert str(u) == ''
        assert u == xml('')
        assert u == xml(None)
        assert u == ''
        assert xml('<') == '<'
        assert xml_quote('<') == '&lt;'
        u += 1
        u += None
        assert u == '1'
        assert xml('a%s') % '&' == 'a&amp;'
        assert xml('a%(a)s%(b)s') % dict(a='&', b=xml('<')) == 'a&amp;<'
        assert xml('a%s%s') % ('&', xml('<')) == 'a&amp;<'
        t = (1,2)
        try:
            t += u
            assert 0
        except TypeError:
            pass
        t = []
        t += u
        assert t == list(u)

    def test_a(self):
        s = xml('<') + '<'
        assert isinstance(s, xml) and s == '<&lt;'
        s = '<' + xml('<')
        assert isinstance(s, xml) and s == '&lt;<', '__radd__ not working.'
        s = xml('%s') % '<'
        assert isinstance(s, xml) and  s == '&lt;'
        s = xml('%s') % xml('<')
        assert isinstance(s, xml) and s == '<'
        s = xml('%(a)s') % dict(a='<')
        assert isinstance(s, xml) and s == '&lt;'
        s = unicode_string('%s') % xml('a')
        assert isinstance(s, unicode_string)
        assert s == 'a'
        s = xml('a')
        s += '<'
        assert s == 'a&lt;'

    def test_quote_wrapper(self):
        assert xml("%0.2f") % 2 == '2.00'
        assert type(xml("%0.2f") % 2) is xml
        assert xml("%(a)s %(b)0.2f") % dict(
            a=unicode_string('<'), b=2) == '&lt; 2.00'
        assert type(xml("%(a)s %(b)0.2f") % dict(
            a=unicode_string('a'), b=2)) is xml
        if str is not unicode_string:
            assert xml("%(a)r %(b)0.2f") % dict(
                a=unicode_string('<'), b=2) == "u&#39;&lt;&#39; 2.00"
        else:
            print(repr(xml("%(a)r %(b)0.2f") % dict(
                a=unicode_string('<'), b=2)))
            assert xml("%(a)r %(b)0.2f") % dict(
                a=unicode_string('<'), b=2) == "&#39;&lt;&#39; 2.00"
        assert type(xml("%(a)r %(b)0.2f")) is xml

    def test_pickle(self):
        from pickle import loads, dumps
        s = xml('hello')
        assert s == loads(dumps(xml(s), 2))

if __name__ == '__main__':
    TestXml()



"""
$URL$
$Id$
"""
from sancho.utest import UTest
from qpy import xml_quote, xml, join_xml, join_str
import sys


class TestQuotedRefs (UTest):

    def refs_quote(self):
        refs = sys.gettotalrefcount
        before = 0
        after = 0
        before = refs()
        xml()
        after = refs()
        assert after - before == 0, after - before
        before = refs()
        xml_quote(None)
        xml_quote("<>'\"")
        after = refs()
        assert after - before == 0, after - before

    def refs_join_str(self):
        refs = sys.gettotalrefcount
        before = 0
        after = 0
        for a in  ([], range(3), 'abc', (xml('<'), 'ok', None, 1)):
            join_str(a) # The refs do change on the first call.
            before = refs()
            join_str(a)
            after = refs()
            assert after - before == 0, after - before

    def refs_join_xml(self):
        refs = sys.gettotalrefcount
        before = 0
        after = 0
        for a in  ([], range(3), 'abc', (xml('<'), 'ok', None, 1)):
            join_xml(a)
            before = refs()
            join_xml(a)
            after = refs()
            assert after - before == 0, after - before

    def a(self):
        emptyrefs = sys.getrefcount(xml())
        refs = sys.gettotalrefcount
        before = 0
        after = 0
        xml('<') + '>'
        xml('<%s>') % '<ok>'
        xml('<%s, %s>') % (1, '<ok>')
        xml('<%(a)s, %(b)s>') % dict(a='<', b=None, c='<')
        r = range(100)
        xml().join(a for a in r)
        x = 0
        before = refs()
        for x in r:
            xml('<') + '>'
            xml('<%s>') % '<ok>'
            xml('<%s, %s>') % (1, '<ok>')
            xml('<%(a)s, %(b)s>') % dict(a='<', b=None, c='<')
            xml().join(a for a in r)
        after = refs()
        assert after - before == 0, after - before
        emptyrefs_after = sys.getrefcount(xml())
        assert emptyrefs_after - emptyrefs == 0, emptyrefs_after

if __name__ == '__main__':
    if hasattr(sys, 'gettotalrefcount'):
        TestQuotedRefs()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/compile.py $
$Id: compile.py 31209 2008-10-13 12:49:03Z dbinger $
"""
from copy import deepcopy
from os import stat, listdir
from os.path import join, splitext
from parser import compilest, sequence2st
from qpy.translate import statement_tree, expr_tree, get_power
from qpy.translate import translate_tokens, get_parse_tree
import py_compile
import symbol
import sys
import sys
import token

if sys.version < "3":
    import __builtin__ as builtins
else:
    import builtins

def get_argument(power):
    assert power[0] == symbol.power
    argument = power[2][2][1]
    assert argument[0] == symbol.argument
    return argument

def is_future_import_statement(node):
    try:
        return (node[0] == symbol.stmt and
            node[1][1][1][1][0] == symbol.import_from and
            node[1][1][1][1][2][1][1] == '__future__')
    except IndexError:
        return False

def is_atom_string(node):
    try:
        power = get_power(node)
        return (len(power) == 2 and
            power[1][0] == symbol.atom and
            power[1][1][0] == token.STRING)
    except IndexError:
        return False

def get_funcdef_function_name_child(node):
    assert node[0] == symbol.funcdef
    for child in node[1:]:
        if child[0] == 1 and child[1] != 'def':
            return child

class Transform (object):

    def __init__(self, tree):
        self.tree = tree
        self.template_type_stack = [None]
        self.module_start = statement_tree(
            'from qpy.quoted import join_xml as _qpy_join_xml, '
            'join_str as _qpy_join_str, xml as _qpy_xml')
        self.template_start = statement_tree(
            'qpy_accumulation=[];qpy_append=qpy_accumulation.append')
        self.return_xml = statement_tree(
            'return _qpy_join_xml(qpy_accumulation)')
        self.return_str = statement_tree(
            'return _qpy_join_str(qpy_accumulation)')
        self.qpy_append_expr_stmt = statement_tree('qpy_append(X)')[1][1][1]
        assert self.qpy_append_expr_stmt[0] == symbol.expr_stmt
        self.xml_power = get_power(expr_tree('_qpy_xml("X")'))
        self.template_type_stack = [None]
        self.traverse_node(self.tree)
        self.line_number = 1
        self.rationalize_line_numbers(self.tree)

    def rationalize_line_numbers(self, node):
        if not isinstance(node, list):
            return
        if len(node) == 3 and node[0] in token.tok_name:
            if node[2] < self.line_number:
                node[2] = self.line_number
            else:
                self.line_number = node[2]
        else:
            for child in node[1:]:
                self.rationalize_line_numbers(child)

    def traverse_node(self, node):
        if not isinstance(node, list):
            return
        # If this is a funcdef, push 'xml', 'str', or None
        if node[0] == symbol.funcdef:
            function_name = get_funcdef_function_name_child(node)[1]
            if function_name.endswith('__xml_template__'):
                self.template_type_stack.append('xml')
            elif function_name.endswith('__str_template__'):
                self.template_type_stack.append('str')
            else:
                self.template_type_stack.append(None)
        # Traverse down before doing modifications.
        for child in node[1:]:
            self.traverse_node(child)
        # Modify the node as necessary.
        if node[0] == symbol.file_input:
            # Insert module-level import statement.
            # Skip over the module docstring and any __future__ imports.
            for index, child in enumerate(node):
                if index == 0:
                    continue
                if is_atom_string(child):
                    continue
                if is_future_import_statement(child):
                    continue
                node.insert(index, deepcopy(self.module_start))
                break
        elif self.template_type_stack[-1] is None:
            pass # We're not in a template, so we're done.
        elif node[0] == symbol.expr_stmt and len(node) == 2:
            # Wrap this expression statement in a qpy_append call.
            stmt = deepcopy(self.qpy_append_expr_stmt)
            argument = get_argument(get_power(stmt))
            assert argument[1][0] == node[1][1][0]
            argument[1] = node[1][1]
            assert node[1][0] == stmt[1][0]
            node[1] = stmt[1]
        elif node[0] == symbol.funcdef:
            # This is a new template.
            # Insert the initialization of qpy_accumulation and qpy_append.
            func_suite = node[-1]
            for j, child in enumerate(func_suite[1:]):
                if child[0] == symbol.stmt:
                    func_suite.insert(j+1, deepcopy(self.template_start))
                    break
            # Add the appropriate return statement and patch function name.
            function_name_node = get_funcdef_function_name_child(node)
            function_name = function_name_node[1]
            if self.template_type_stack[-1] == 'xml':
                return_accumulation = deepcopy(self.return_xml)
                # trim __xml_template__
                function_name_node[1] = function_name_node[1][:-16] 
            else:
                assert self.template_type_stack[-1] == 'str'
                return_accumulation = deepcopy(self.return_str)
                # trim __str_template__                
                function_name_node[1] = function_name_node[1][:-16] 
            func_suite.insert(-1, return_accumulation)
        elif (self.template_type_stack[-1] == 'xml' and
            node[0] == symbol.power and
            node[1][0] == symbol.atom and
            node[1][1][0] == token.STRING):
            # node looks like
            # [power [atom [STRING "Z" 1]] ...]
            xml_power = deepcopy(self.xml_power)
            # xml_power looks like
            # [power
            #     [atom [STRING "_qpy_xml" 1]
            #     [trailer 
            #         LPAR ... 
            #         [arglist ... [power [atom [STRING "X"]]]]
            #         RPAR]]]
            argument = get_argument(xml_power)
            argument_power = get_power(argument)
            assert argument_power[1][0] == symbol.atom
            assert argument_power[1][0] == node[1][0]
            argument_power[1] = deepcopy(node[1]) # replace "X"
            node[1] = xml_power[1]
            node.insert(2, xml_power[2])
        # Pop the stack.
        if node[0] == symbol.funcdef:
            self.template_type_stack.pop()

    def get_st(self):
        return sequence2st(self.tree)

def get_code(source, source_name):
    translated_source = translate_tokens(source)
    tree = get_parse_tree(translated_source, source_name)
    transformed = Transform(tree).get_st()
    code = compilest(transformed, filename=source_name)
    return code

def timestamp(filename):
    try:
        s = stat(filename)
    except OSError:
        return None
    return s.st_mtime

def compile_qpy_file(source_name):
    """(source_name:str)
    Compile the given filename if it is a .qpy file and if it does not already
    have an up-to-date .pyc file.
    """
    if source_name[-4:] == '.qpy':
        compile_time = timestamp(source_name[:-4] + '.pyc')
        if compile_time is not None:
            source_time = timestamp(source_name)
            if compile_time >= source_time:
                return # already up-to-date
        compile(source_name)

def compile_qpy_files(path):
    """(path:str)
    Compile the .qpy files in the given directory.
    """
    for name in listdir(path):
        if name[-4:] == '.qpy':
            compile_qpy_file(join(path, name))

def compile(source_name):
    # Replace builtins.compile (temporarily) and use the py_compile module
    # to create the .pyc file using our special compile() function.
    output_name = splitext(source_name)[0] + ".pyc"
    builtins_compile = builtins.compile
    try:
        def qpycompile(source, source_name, extra='exec'):
            assert extra == 'exec'
            return get_code(source, source_name)
        builtins.compile = qpycompile
        py_compile.compile(source_name, cfile=output_name, doraise=True)
    finally:
        builtins.compile = builtins_compile

if __name__ == '__main__':
    import qpy.__main__



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/quoted.py $
$Id: quoted.py 31195 2008-10-04 19:09:10Z dbinger $
"""
import sys

if not sys.version.startswith('3'):
    from __builtin__ import unicode, basestring, long # for checker.
    unicode_str = unicode
    number_classes = (int, long, float)
    string = basestring
else:
    number_classes = (int, float)
    string = str
    unicode_str = str

def xml_quote(x):
    """(x:anything) -> xml
    """
    if isinstance(x, xml):
        return x
    if x is None:
        return xml()
    if type(x) is unicode_str:
        s = x
    elif isinstance(x, string):
        s = unicode_str(x, 'utf-8')
    else:
        s = unicode_str(x)
    s = s.replace("&", "&amp;")
    s = s.replace("<", "&lt;")
    s = s.replace(">", "&gt;")
    s = s.replace('"', "&quot;")
    s = s.replace("'", "&#39;") # &apos; is not HTML.
    return xml(s)

class _xml_quote_wrapper (object):
    """
    Not for outside code.
    For values that are used as arguments to the % operator, this allows
    str(value) and repr(value), if called as part of the formatting, to
    produce quoted results.
    """
    __slots__ = ['value']

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return xml_quote(self.value)

    def __repr__(self):
        return xml_quote(repr(self.value))

    def __getitem__(self, key):
        return _quote_wrap(self.value[key])

def _quote_wrap(x):
    """(x) -> _xml_quote_wrapper | x
    Not for outside code.
    Return a value v for which str(v) and repr(v) will produce quoted strings,
    or, if x is a dict, for which str(v[key]) and repr(v[key]) will produce
    quoted strings.
    """
    if isinstance(x, number_classes):
        return x
    else:
        return _xml_quote_wrapper(x)

class xml (unicode_str):
    """
    This subclass of str is designated as needing no more xml quoting.
    """
    __slots__ = []

    def __new__(klass, string=None, encoding=sys.getdefaultencoding(), 
        errors='strict'):
        if string is None:
            return xml('')
        elif isinstance(string, unicode_str):
            string = string.encode('utf-8')
            encoding = 'utf-8'
        return unicode_str.__new__(klass, string, encoding, errors)

    def __add__(self, other):
        """(other) -> xml
        """
        return xml(unicode_str.__add__(self, xml_quote(other)))

    def __radd__(self, other):
        """(other:str) -> xml
        """
        if isinstance(other, string):
            return xml(unicode_str.__add__(xml_quote(other), self))
        else:
            return NotImplemented

    def __mul__(self, other):
        """(other) -> xml
        """
        return xml(unicode_str.__mul__(self, other))

    def __rmul__(self, other):
        """(other) -> xml
        """
        return xml(unicode_str.__mul__(self, other))

    def __mod__(self, other):
        """(other) -> xml
        """
        if isinstance(other, tuple):
            target = tuple(_quote_wrap(item) for item in other)
        else:
            target = _quote_wrap(other)
        return xml(unicode_str.__mod__(self, target))

    def join(self, items):
        """(items) -> xml
        """
        return xml(unicode_str.join(self, (xml_quote(item) for item in items)))

join_xml = xml().join
xml.quote = staticmethod(xml_quote)

def join_str(args):
    return ''.join(unicode_str(arg) for arg in args if arg is not None)

__all__ = ['xml', 'join_xml', 'join_str', 'xml_quote']



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/__init__.py $
$Id: __init__.py 31745 2009-07-20 21:21:48Z dbinger $
Copyright (c) Corporation for National Research Initiatives 2009. All Rights Reserved.
"""
__version__ = "1.8"

import sys

try:
    from qpy.quoted import xml, join_str, join_xml, xml_quote
    [xml, join_str, join_xml, xml_quote] # for checker.
except ImportError:
    sys.stdout.write("qpy %s\n" % __version__)
    xml = None

# The remainder is for partial compatibility with previous release.
h8 = xml
import sys
if sys.version.startswith("3"):
    stringify = str
    u8 = str
else:
    from __builtin__ import unicode, basestring # for checker.
    u8 = unicode

    def stringify(obj):
        """(obj) -> basestring
        Return a string version of `obj`.  This is like str(), except
        that it tries to prevent turning str instances into unicode instances.
        The type of the result is either str or unicode.
        """
        tp = type(obj)
        if issubclass(tp, basestring):
            return obj
        elif hasattr(tp, '__unicode__'):
            s = tp.__unicode__(obj)
            if not isinstance(s, basestring):
                raise TypeError('__unicode__ did not return a string')
            return s
        elif hasattr(tp, '__str__'):
            s = tp.__str__(obj)
            if not isinstance(s, basestring):
                raise TypeError('__str__ did not return a string')
            return s
        else:
            return str(obj)



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/qpcheck.py $
$Id: qpcheck.py 31746 2009-07-21 13:03:12Z dbinger $
"""
from qpy.translate import get_parse_tree, symtree, translate_tokens
import symbol
import token
import sys
import keyword

python2 = not sys.version.startswith("3")

if python2:
    import __builtin__ as builtins
else:
    import builtins

def p(*args):
    for arg in args[:-1]:
        sys.stdout.write(str(arg) + " ")
    sys.stdout.write(str(args[-1]) + "\n")


class Namespace (object):

    def __init__(self, name, previous):
        self.names = set()
        self.name = name
        self.previous = previous

    def add(self, name):
        self.names.add(name)

    def update(self, names):
        for name in names:
            self.add(name)

    def __contains__(self, name):
        return name in self.names or name in self.previous

    def __str__(self):
        if not self.previous or not self.previous.get_previous():
            return self.name
        else:
            return str(self.previous) + "." + self.name

    def get_names(self):
        return self.names

    def get_base(self):
        if self.previous == []:
            return self
        else:
            return self.previous.get_base()

    def get_previous(self):
        return self.previous

def get_tree(filename):
    source = open(filename).read()
    if filename.endswith('.qpy'):
        source = translate_tokens(source)
    return get_parse_tree(source, filename)

def get_global_namespace():
    namespace = Namespace('GLOBAL', [])
    namespace.update(keyword.kwlist)
    namespace.add('__builtins__')
    namespace.add('__doc__')
    namespace.add('__file__')
    namespace.add('__name__')
    namespace.add('__path__')
    namespace.update(dir(builtins))
    return namespace

def get_node_name(node):
    number = node[0]
    return symbol.sym_name.get(number) or token.tok_name.get(number)

def get_name(node):
    if node[0] == token.NAME:
        return node[1]
    else:
        return None


def get_arglist_names(node):
    names = []
    node_name = get_node_name(node)
    if node_name not in ('typedargslist', 'varargslist'):
        return names
    if python2:
        defnames = ('fpdef', 'fplist')
    else:
        defnames =('tfpdef', 'vfpdef')
    todo = node[1:]
    while todo:
        child = todo[0]
        todo = todo[1:]
        if get_node_name(child) in defnames:
            todo += child[1:]
        if get_node_name(child) == 'NAME':
            name = get_name(child)
            names.append(name)
    return names

def gen_power(node):
    if isinstance(node, list):
        if node[0] == symbol.power:
            yield node
        for n in node[1:]:
            for p in gen_power(n):
                yield p

def gen_names(exprlist):
    # factor: ('+'|'-'|'~') factor | power
    # power: atom trailer* ['**' factor]
    # atom: ('(' [yield_expr|testlist_comp] ')' |
    #        '[' [testlist_comp] ']' |
    #        '{' [dictorsetmaker] '}' |
    #        NAME | NUMBER | STRING+ | '...' | 'None' | 'True' | 'False')
    # testlist_comp: test ( comp_for | (',' test)* [','] )
    # trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME
    for power in gen_power(exprlist):
        if len(power) == 2: # Just an atom
            atom = power[1]
            if atom[1][0] == token.NAME:
                yield get_name(atom[1])


def traverse(node, namespace, unresolved, unused):
    if not isinstance(node, list):
        return
    node_name = get_node_name(node)
    children  = node[1:]
    num_children = len(children)

    if node_name == 'NAME':
        name = get_name(node)
        if name not in namespace and '*' not in namespace:
            line_number = node[-1]
            unresolved[name] = (line_number, name, namespace)
        else:
            unused.discard(name)
        return

    if node_name == 'trailer' and children[0][0] == token.DOT:
        return

    # file_input: (NEWLINE | stmt)* ENDMARKER
    # eval_input: testlist NEWLINE* ENDMARKER
    # decorator: '@' dotted_name [ '(' [arglist] ')' ] NEWLINE
    # decorators: decorator+
    # decorated: decorators (classdef | funcdef)
    # funcdef: 'def' NAME parameters ['->' test] ':' suite
    if node_name == 'funcdef':
        if python2:
            name = get_name(children[-4])
            parameters = children[-3]
        else:
            name = get_name(children[1])
            parameters = children[2]

        if name.endswith('__xml_template__'):
            name = name[:-16]
        if name.endswith('__str_template__'):
            name = name[:-16]

        suite = children[-1]
        if get_node_name(children[-3]) != 'parameters':
            result_test = children[4]
            traverse(result_test, namespace, unresolved, unused)
        namespace = Namespace(name, namespace)
        traverse(parameters, namespace, unresolved, unused)
        namespace.get_previous().add(name)
        arg_names = get_arglist_names(parameters[-2])
        namespace.update(arg_names)
        traverse(suite, namespace, unresolved, unused)
        namespace = namespace.get_previous()
        return

    # parameters: '(' [typedargslist] ')'
    # typedargslist: ((tfpdef ['=' test] ',')*
    #                 ('*' [tfpdef] (',' tfpdef ['=' test])* [',' '**' tfpdef] | '**' tfpdef)
    #                 | tfpdef ['=' test] (',' tfpdef ['=' test])* [','])

    # tfpdef: NAME [':' test]
    if node_name == 'tfpdef':
        # Don't worry about resolving the parameter name.
        for child in children[1:]:
            traverse(child, namespace, unresolved, unused)
        return

    # varargslist: ((vfpdef ['=' test] ',')*
    #               ('*' [vfpdef] (',' vfpdef ['=' test])*  [',' '**' vfpdef] | '**' vfpdef)
    #               | vfpdef ['=' test] (',' vfpdef ['=' test])* [','])

    if node_name == 'varargslist':
        for j in range(1, len(children)):
            if get_node_name(children[j-1]) in ('STAR', 'DOUBLESTAR'):
                name = get_name(children[j])
                if name:
                    namespace.add(name)

    # vfpdef: NAME
    if node_name == 'vfpdef':
        # Don't worry about resolving the parameter name.
        return

    if python2 and node_name == 'fpdef':
        # Don't worry about resolving the parameter name.
        return

    # stmt: simple_stmt | compound_stmt
    # simple_stmt: small_stmt (';' small_stmt)* [';'] NEWLINE
    # small_stmt: (expr_stmt | del_stmt | pass_stmt | flow_stmt |
    #              import_stmt | global_stmt | nonlocal_stmt | assert_stmt)

    # expr_stmt: testlist (augassign (yield_expr|testlist) |
    #                      ('=' (yield_expr|testlist))*)
    # You can't really assign to a yield expr.
    # The testlists must also be some kind of names.
    if node_name == 'expr_stmt':
        if num_children >= 3 and children[1][0] == token.EQUAL:
            traverse(node[-1], namespace, unresolved, unused)
            for child in children[:-2:2]:
                if child[0] == symbol.testlist:
                    names = list(gen_names(child))
                    namespace.update(names)
                else:
                    assert 0, symtree(node)

    # augassign: ('+=' | '-=' | '*=' | '/=' | '%=' | '&=' | '|=' | '^=' |
    #             '<<=' | '>>=' | '**=' | '//=')
    # # For normal assignments, additional restrictions enforced by the interpreter
    # del_stmt: 'del' exprlist

    # pass_stmt: 'pass'
    # flow_stmt: break_stmt | continue_stmt | return_stmt | raise_stmt | yield_stmt
    # break_stmt: 'break'
    # continue_stmt: 'continue'
    # return_stmt: 'return' [testlist]
    # yield_stmt: yield_expr
    # raise_stmt: 'raise' [test ['from' test]]
    # import_stmt: import_name | import_from
    # import_name: 'import' dotted_as_names

    # # note below: the ('.' | '...') is necessary because '...' is tokenized as ELLIPSIS
    # import_from: ('from' (('.' | '...')* dotted_name | ('.' | '...')+)
    #               'import' ('*' | '(' import_as_names ')' | import_as_names))

    if node_name == 'import_from':
        last_child = children[-1]
        if last_child[0] == token.STAR:
            namespace.add('*')
        else:
            if last_child[0] == symbol.import_as_names:
                import_as_names = last_child
            else:
                import_as_names = children[-2]
            assert import_as_names[0] == symbol.import_as_names
            for import_as_name in import_as_names[1::2]: # skip commas
                name = get_name(import_as_name[-1])
                namespace.add(name)
                unused.add(name)
        return

    # import_as_name: NAME ['as' NAME]
    if node_name == 'import_as_name':
        name = get_name(children[-1])
        namespace.add(name)
        return

    # dotted_as_name: dotted_name ['as' NAME]
    if node_name == 'dotted_as_name':
        if num_children == 1:
            dotted_name = children[0]
            names = []
            for n in dotted_name[1::2]:
                names.append(get_name(n))
                namespace.add('.'.join(names))
        else:
            name = get_name(children[-1])
            namespace.add(name)
        return

    # import_as_names: import_as_name (',' import_as_name)* [',']
    # dotted_as_names: dotted_as_name (',' dotted_as_name)*
    # dotted_name: NAME ('.' NAME)*
    if node_name == 'dotted_name':
        name = '.'.join(get_name(n) for n in node[1::2])
        return

    # global_stmt: 'global' NAME (',' NAME)*
    # nonlocal_stmt: 'nonlocal' NAME (',' NAME)*
    # assert_stmt: 'assert' test [',' test]
    #
    # compound_stmt: if_stmt | while_stmt | for_stmt | try_stmt | with_stmt | funcdef | classdef | decorated
    # if_stmt: 'if' test ':' suite ('elif' test ':' suite)* ['else' ':' suite]
    # while_stmt: 'while' test ':' suite ['else' ':' suite]

    # for_stmt: 'for' exprlist 'in' testlist ':' suite ['else' ':' suite]
    if node_name == 'for_stmt':
        exprlist = children[1]
        names = list(gen_names(exprlist))
        namespace.update(names)

    # try_stmt: ('try' ':' suite
    #            ((except_clause ':' suite)+
    #       ['else' ':' suite]
    #       ['finally' ':' suite] |
    #      'finally' ':' suite))
    # with_stmt: 'with' test [ with_var ] ':' suite

    # with_var: 'as' expr # python2.6
    # with_item: 'as' expr # python3.2

    if node_name in ('with_var', 'with_item'):
        exprlist = children[-1]
        names = list(gen_names(exprlist))
        if names:
            namespace.update(names)

    # # NB compile.c makes sure that the default except clause is last
    # except_clause: 'except' [test ['as' NAME]]

    if node_name == 'except_clause':
        if num_children > 1:
            traverse(children[1], namespace, unresolved, unused) # test
        if num_children > 2:
            if get_node_name(children[-1]) == 'NAME':
                name = get_name(children[-1])
                namespace.add(name)
            if python2 and get_node_name(children[-1]) == 'test':
                names = list(gen_names(children[-1]))
                namespace.update(names)
        return

    # suite: simple_stmt | NEWLINE INDENT stmt+ DEDENT
    #
    # test: or_test ['if' or_test 'else' test] | lambdef
    # test_nocond: or_test | lambdef_nocond
    # lambdef: 'lambda' [varargslist] ':' test
    # lambdef_nocond: 'lambda' [varargslist] ':' test_nocond
    if node_name in ('lambdef', 'lambdef_nocond'):
        namespace = Namespace('<lambda>', namespace)
        if num_children == 4:
            parameters = children[1]
            traverse(parameters, namespace, unresolved, unused)
            arg_names = get_arglist_names(parameters)
            namespace.update(arg_names)
        traverse(children[-1], namespace, unresolved, unused)
        namespace = namespace.get_previous()
        return

    # or_test: and_test ('or' and_test)*
    # and_test: not_test ('and' not_test)*
    # not_test: 'not' not_test | comparison
    # comparison: star_expr (comp_op star_expr)*
    # comp_op: '<'|'>'|'=='|'>='|'<='|'!='|'in'|'not' 'in'|'is'|'is' 'not'
    # star_expr: ['*'] expr
    # expr: xor_expr ('|' xor_expr)*
    # xor_expr: and_expr ('^' and_expr)*
    # and_expr: shift_expr ('&' shift_expr)*
    # shift_expr: arith_expr (('<<'|'>>') arith_expr)*
    # arith_expr: term (('+'|'-') term)*
    # term: factor (('*'|'/'|'%'|'//') factor)*
    # factor: ('+'|'-'|'~') factor | power
    # power: atom trailer* ['**' factor]
    # atom: ('(' [yield_expr|testlist_comp] ')' |
    #        '[' [testlist_comp] ']' |
    #        '{' [dictorsetmaker] '}' |
    #        NAME | NUMBER | STRING+ | '...' | 'None' | 'True' | 'False')

    # testlist_comp: test ( comp_for | (',' test)* [','] )
    if node_name == 'testlist_comp':
        if num_children == 2 and get_node_name(children[1]) == 'comp_for':
            traverse(children[1], namespace, unresolved, unused)
            for child in children:
                if child is not children[1]:
                    traverse(child, namespace, unresolved, unused)
            return

    # trailer: '(' [arglist] ')' | '[' subscriptlist ']' | '.' NAME
    # subscriptlist: subscript (',' subscript)* [',']
    # subscript: test | [test] ':' [test] [sliceop]
    # sliceop: ':' [test]
    # exprlist: star_expr (',' star_expr)* [',']
    # testlist: test (',' test)* [',']
    # dictorsetmaker: ( (test ':' test (comp_for | (',' test ':' test)* [','])) |
    #                   (test (comp_for | (',' test)* [','])) )
    #

    # classdef: 'class' NAME ['(' [arglist] ')'] ':' suite
    if node_name == 'classdef':
        name = get_name(children[1])
        for n in children[2:-2]:
            traverse(n, namespace, unresolved, unused)
        namespace.add(name)  # a little too early
        namespace = Namespace(name, namespace)
        traverse(children[-1], namespace, unresolved, unused) # suite
        namespace = namespace.get_previous()
        return

    #
    # arglist: (argument ',')* (argument [',']| '*' test [',' '**' test] | '**' test)
    # argument: test [comp_for] | test '=' test  # Really [keyword '='] test
    if node_name == 'argument':
        if num_children == 3:
            traverse(children[-1], namespace, unresolved, unused)
            return
        if num_children == 2:
            childname = get_node_name(children[1])
            if childname == 'comp_for' or (python2 and childname == 'gen_for'):
                traverse(children[1], namespace, unresolved, unused)

    if python2 and node_name == 'testlist_gexp':
        for child in children:
            if get_node_name(child) == 'gen_for':
                traverse(child, namespace, unresolved, unused)

    if python2 and node_name in 'listmaker':
        for child in children:
            if get_node_name(child) == 'list_for':
                traverse(child, namespace, unresolved, unused)
    #
    # comp_iter: comp_for | comp_if
    # comp_for: 'for' exprlist 'in' or_test [comp_iter]
    if (node_name == 'comp_for' or
        (python2 and node_name in ('gen_for', 'list_for'))):
        exprlist = children[1]
        names = list(gen_names(exprlist))
        assert names
        namespace.update(names)

    # comp_if: 'if' test_nocond [comp_iter]
    #
    # testlist1: test (',' test)*

    # The default traversal of children
    for child in children:
        traverse(child, namespace, unresolved, unused)


def check(filename):
    global_namespace = get_global_namespace()
    try:
        tree = get_tree(filename)
    except SyntaxError:
        e = sys.exc_info()[1]
        lineno = e.lineno
        f = open(filename)
        line = 0
        for x in range(lineno):
            line = f.readline()
        f.close()
        p(filename, "line %s SyntaxError %r" % (e.lineno, line))
        return
    unresolved = dict()
    unused = set()
    traverse(tree, global_namespace, unresolved, unused)
    for line_number, name, namespace in sorted(unresolved.values()):
        if name not in global_namespace:
            msg = '(%s, in %s): %r is undefined.' % (line_number, namespace, name)
            p(filename, msg)
    for name in unused:
        if not name.startswith('_'):
            p("%s: %r is unused" % (filename, name))



def main():
    from optparse import OptionParser
    from os import listdir
    from os.path import join, isdir, isfile, basename

    parser = OptionParser()
    parser.set_description(
        'Check python source files for unknown name errors '
        'and for unused imports. '
        'The arguments name files to check or '
        'directories to check recursively. '
        'If no arguments are given, the current directory '
        'is checked. '
        'If there is a recursion, it excludes ".svn", "build", '
        'and "dist" directories, and includes files that end '
        'in ".py" or ".qpy".')
    (options, args) = parser.parse_args()
    if args:
        todo = args
    else:
        todo = ['.']
    while todo:
        arg = todo.pop()
        if basename(arg) in ['.svn', 'dist', 'build']:
            continue
        elif (isfile(arg) and
              arg in sys.argv or
              (arg.endswith('.py') or
               arg.endswith('.qpy'))):
            check(arg)
        elif isdir(arg):
            todo.extend([join(arg, item) for item in listdir(arg)])

if __name__ == '__main__':
    main()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/test/utest_translate.py $
$Id: utest_translate.py 30201 2007-11-02 15:25:24Z dbinger $
"""
from sancho.utest import UTest, raises
from qpy.translate import translate_tokens, get_parse_tree, ptree, symtree
from qpy.translate import statement_tree, expr_tree


class TestTranslate(UTest):

    def test_translate(self):
        src = (
            'def a:xml():\n'
            '    "ok"\n')
        translated = translate_tokens(src)
        assert 'def a__xml_template__(' in translated, translated

        src = ('def a [html] ():\n'
               '    "ok"\n')
        assert translated == translate_tokens(src)
        assert 'def a__xml_template__(' in translated

        src = (
            'def a:str():\n'
            '    "ok"\n')
        translated = translate_tokens(src)
        assert 'def a__str_template__(' in translated

        src = ('def a [plain] ():\n'
               '    "ok"\n')
        assert translated == translate_tokens(src)
        assert 'def a__str_template__(' in translated

    def test_tree_utils(self):
        src = ('def a():\n'
               '    "ok"\n')
        tree = get_parse_tree(src, 'fake')
        assert type(tree) is list
        raises(SyntaxError, get_parse_tree, 'bogus' + src)
        raises(SyntaxError, get_parse_tree, 'bogus' + src, 'fake')
        statement_tree(src)
        expr_tree('x+1')
        symtree(tree)
        ptree(tree)
        raises(AssertionError, symtree, [-1])

if __name__ == '__main__':
    TestTranslate()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/__main__.py $
$Id: __main__.py 30591 2008-03-13 15:39:38Z dbinger $
"""
from qpy.compile import compile
import sys

args = sys.argv[1:]
if not args:
    sys.stdout.write("no files to compile\n")
else:
    for filename in args:
        compile(filename)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/setup.py $
$Id: setup.py 31598 2009-04-29 18:51:25Z dbinger $
"""
import sys
assert sys.version >= "2.4"
sys.path.insert(0, '..')
from glob import glob
import re, os

try:
    assert 'USE_DISTUTILS' not in os.environ
    from setuptools import setup, Extension
except (ImportError, AssertionError):
    from distutils.core import setup
    from distutils.extension import Extension

from distutils.command.build_py import build_py


class qpy_build_py (build_py):

    def find_package_modules(self, package, package_dir):
        self.check_package(package, package_dir)
        module_files = (glob(os.path.join(package_dir, "*.py")) +
                        glob(os.path.join(package_dir, "*.qpy")))
        modules = []
        setup_script = os.path.abspath(self.distribution.script_name)
        for f in module_files:
            abs_f = os.path.abspath(f)
            module = os.path.splitext(os.path.basename(f))[0]
            modules.append((package, module, f))
        return modules

    def build_module(self, module, module_file, package):
        if type(package) is str:
            package = package.split('.')
        elif type(package) not in (list, tuple):
            raise TypeError(
                "'package' must be a string (dot-separated), list, or tuple")
        # Now put the module source file into the "build" area.
        outfile = self.get_module_outfile(self.build_lib, package, module)
        if module_file.endswith(".qpy"):
            outfile = outfile[0:outfile.rfind('.')] + ".qpy"
        dir = os.path.dirname(outfile)
        self.mkpath(dir)
        return self.copy_file(module_file, outfile, preserve_mode=0)

    def byte_compile(self, files):
        try:
            from qpy.compile import compile_qpy_file
        except ImportError:
            build_py.byte_compile(self, files)
        else:
            existing = []
            for file in files:
                if os.path.exists(file):
                    existing.append(file)
                else:
                    alt = file[:-3] + ".qpy"
                    compile_qpy_file(alt)
                    sys.stdout.write('byte-compiling %s\n' % alt)
            build_py.byte_compile(self, existing)


if __name__ == '__main__':
    from __init__ import __version__
    if 'sdist' in sys.argv:
        if sys.platform == 'darwin':
            # Omit extended attributes from tarfile
            os.environ['COPYFILE_DISABLE'] = 'true'
        import re
        # Make sure that version numbers have all been updated.
        PAT = re.compile(r'\b%s\b' % re.escape(__version__))
        assert len(PAT.findall(open("LICENSE.txt").read())) == 14, __version__
        assert PAT.search(open("CHANGES.txt").readline()), __version__
        assert len(PAT.findall(open("README.txt").read())) == 0, __version__

        # Make sure that copyright statements are current.
        from datetime import datetime
        year = datetime.now().year # no tz, ok
        copyright = "Copyright (c) Corporation for National Research Initiatives %s" % year
        assert open("__init__.py").read().count(copyright) == 1
        assert open("README.txt").read().count(copyright) == 1
    ext_modules = [Extension(name="qpy.quoted", sources=["quoted.c"])]
    setup(
        name="qpy",
        version=__version__,
        description="xml/html content in functions",
        author="CNRI",
        author_email="webmaster@mems-exchange.org",
        url="http://www.mems-exchange.org/software/qpy/",
        license="see LICENSE.txt",
        package_dir=dict(qpy=os.curdir),
        packages=['qpy'],
        scripts=['qpcheck.py', 'qpyrun.py'],
        ext_modules=ext_modules,
        cmdclass= {'build_py': qpy_build_py},
        long_description=(
        "Qpy provides a convenient mechanism for generating safely-quoted xml "
        "from python code.  It does this by implementing a quote-no-more string "
        "data type and a slightly modified python compiler."
        ))



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/qpy/translate.py $
$Id: translate.py 30587 2008-03-13 10:55:08Z dbinger $
"""
from parser import expr, suite
from pprint import pprint
from symbol import sym_name
from token import tok_name
import re
import sys
import symbol

_annotation_re = re.compile(
    r"^(?P<indent>[ \t]*)def(?:[ \t]+)"
    r"(?P<name>[a-zA-Z_][a-zA-Z_0-9]*)"
    r"(?:[ \t]*:[ \t]*)(?P<type>xml|str)(?:[ \t]*)"
    r"(?:[ \t]*[\(\\])",
    re.MULTILINE|re.VERBOSE)

_template_re = re.compile(
    r"^(?P<indent>[ \t]*) def (?:[ \t]+)"
    r" (?P<name>[a-zA-Z_][a-zA-Z_0-9]*)"
    r" (?:[ \t]*) \[(?P<type>plain|html)\] (?:[ \t]*)"
    r" (?:[ \t]*[\(\\])",
    re.MULTILINE|re.VERBOSE)

def translate_tokens(buf):
    """
    def f:xml( ... )  -> def f__xml_template__(...):
    def f:str( ... )  -> def f__str_template__(...):

    and, for backward compatibility,

    def foo [html] (...): -> def foo__xml_template__(...):
    def foo [plain] (...): -> def foo__str_template__(...):
    """
    def replacement(match):
        if match.group('type') in ('xml', 'html'):
            template_type = 'xml'
        elif match.group('type') in ('str', 'plain'):
            template_type = 'str'
        return '%sdef %s__%s_template__(' % (match.group('indent'),
                                 match.group('name'),
                                 template_type)
    translated = _annotation_re.sub(replacement, buf)
    translated = _template_re.sub(replacement, translated)    
    return translated

def get_parse_tree(source, source_name=None):
    try:
        if source_name is None:
            return expr(source).tolist(True)
        else:
            return suite(source).tolist(True)
    except (TypeError, SyntaxError):
        e = sys.exc_info()[1]
        e.filename = source_name
        raise e

def statement_tree(stmt):
    tree = get_parse_tree(stmt, 'fake')[1]
    assert tree[0] == symbol.stmt
    return tree

def expr_tree(x):
    tree = get_parse_tree(x)[1]
    assert tree[0] == symbol.testlist
    return tree

def get_power(node):
    n = node
    while n[0] != symbol.power:
        n = n[1]
    return n

def symtree(t):
    if not isinstance(t, list):
        return t
    if t[0] in sym_name:
        return [(sym_name.get(t[0]), t[0])] + list(map(symtree, t[1:]))
    elif t[0] in tok_name:
        return [(tok_name.get(t[0]), t[0])] + list(map(symtree, t[1:]))
    else:
        assert 0, t

def ptree(t):
    pprint(symtree(t))
